module.exports = grammar({
  name: "BorneoScript",

  rules: {
    //Define characters
    comma: $ => choice($.automatic_comma, ","),

    //Define keywords
    statement: $ => choice(
      $.if_statement
    ),

    //Basic grammar
    assignment_pattern: $ => seq(
      field('left', $.pattern),
      optional("=", ":"),
      field('right', $.logic_symbols)
    ),
    block: $ => seq(
      "{",
        repeat($.statement),
      "}"
    ),

    //Comments
    comment: $ => token(choice(
      seq("//", /.*/),
      seq("#", /.*/),
      seq(
        "/*",
        /[^*]*\*+([^/*][^*]*\*+)*/,
        "/"
      )
    )),

    //Conditionals
    logic_symbols: $ => choice(
      ...[
        ["and", "logical_and"],
        ["AND", "logical_and"],
        ["or", "logical_or"],
        ["OR", "logical_or"],
        ["not", "logical_not"],
        ["NOT", "logical_not"]
      ].map([operator, precedence] => {
        prec.left(precedence, seq(
          field("operator", operator)
        ))
      })
    ),

    else_statement: $ => prec.right(seq(
      "else",
      "{",
        repeat($.statement),
      "}"
    )),
    if_statement: $ => prec.right(seq(
      choice("if", "IF"),
      "{"
        field("condition", $.limit_statement),
        field("else", $.else_statement)
      "}"
    )),
    limit_statement: $ => prec.right(seq(
      "limit",
      "{",
        repeat($.statement),
      "}"
    )),

    //Variable types - Booleans, Numbers, Strings, and Arrays
    array: $ => seq(
      "[",
        commaSep(optional(choice(
          $.pattern,
          $.assignment_pattern //Undefined
        ))),
      "]"
    ),
    destructuring_pattern: $ => choice(
      $.array,
      $.object_pattern
    ),
    false: $ => "false",
    escape_sequence: $ => token.immediate(seq(
      '\\',
      choice(
        /[^xu0-7]/,
        /[0-7]{1,3}/,
        /x[0-9a-fA-F]{2}/,
        /u[0-9a-fA-F]{4}/,
        /u{[0-9a-fA-F]+}/
      )
    )),
    identifier: $ => {
      const alpha = /[^\x00-\x1F\s0-9:;`"'@#.,|^&<=>+\-*/\\%?!~()\[\]{}\uFEFF\u2060\u200B\u00A0]|\\u[0-9a-fA-F]{4}|\\u\{[0-9a-fA-F]+\}/
      const alphanumeric = /[^\x00-\x1F\s:;`"'@#.,|^&<=>+\-*/\\%?!~()\[\]{}\uFEFF\u2060\u200B\u00A0]|\\u[0-9a-fA-F]{4}|\\u\{[0-9a-fA-F]+\}/
      return token(seq(alpha, repeat(alphanumeric)))
    },
    null: $ => "null",
    number: $ => { //https://github.com/tree-sitter/tree-sitter-javascript/blob/master/grammar.js
      const hex_literal = seq(
        choice('0x', '0X'),
        /[\da-fA-F](_?[\da-fA-F])*/
      )

      const decimal_digits = /\d(_?\d)*/
      const signed_integer = seq(optional(choice('-', '+')), decimal_digits)
      const exponent_part = seq(choice('e', 'E'), signed_integer)

      const binary_literal = seq(choice('0b', '0B'), /[0-1](_?[0-1])*/)

      const octal_literal = seq(choice('0o', '0O'), /[0-7](_?[0-7])*/)

      const bigint_literal = seq(choice(hex_literal, binary_literal, octal_literal, decimal_digits), 'n')

      const decimal_integer_literal = choice(
        '0',
        seq(optional('0'), /[1-9]/, optional(seq(optional('_'), decimal_digits)))
      )

      const decimal_literal = choice(
        seq(decimal_integer_literal, '.', optional(decimal_digits), optional(exponent_part)),
        seq('.', decimal_digits, optional(exponent_part)),
        seq(decimal_integer_literal, exponent_part),
        seq(decimal_digits),
      )

      return token(choice(
        hex_literal,
        decimal_literal,
        binary_literal,
        octal_literal,
        bigint_literal,
      ))
    },
    object: $ => {
      field("key", $.property_name),
      ":",
      field("value", choice($.logic_symbols))
    },
    object_assignment_pattern: $ => seq(
      field('left', choice(
        alias(choice($.reserved_identifier, $.identifier), $.shorthand_property_identifier_pattern),
        $.destructuring_pattern
      )),
      choice("=", ":"),
      field('right', $.logic_symbols)
    ),
    object_pattern: $ => prec('object', seq(
      '{',
        commaSep(optional(choice(
          $.object,
          $.object_assignment_pattern,
          alias(
            choice($.identifier, $.reserved_identifier),
            $.shorthand_property_identifier_pattern
          )
        ))),
      '}'
    )),
    pattern: $ => prec.dynamic(-1, choice(
      $.identifier,
      alias($.reserved_identifier, $.identifier)
    )),
    string: $ => choice(
      seq(
        '"',
          repeat(choice(
            alias($.unescaped_double_string_fragment, $.string_fragment),
            $.escape_sequence
          )),
        '"'
      ),
      seq(
        "'",
          repeat(choice(
            alias: ($.unescaped_single_string_fragment, $.string_fragment),
            $.escape_sequence
          )),
        "'"
      )
    ),
    property_name: $ => choice(
      $.false,
      $.null,
      $.number,
      $.string,
      $.true,
      $.undefined
    ),
    reserved_identifier: $ => choice(
      'get',
      'set',
      'async',
      'static',
      'export'
    ),
    true: $ => "true",
    undefined: $ => "undefined",
    unescaped_double_string_fragment: $ => token.immediate(prec(1, /[^"\\]+/)),
    unescaped_single_string_fragment: $ => token.immediate(prec(1, /[^'\\]+/)),
  }
});

//Leftovers from JS
function commaSep1(rule) {
  return seq(rule, repeat(seq(',', rule)));
}

function commaSep(rule) {
  return optional(commaSep1(rule));
}
